package br.usp.ime.mac499.xba.preprocess;

import java.io.IOException;
import java.io.Reader;
import java.util.Stack;


/**
 * Responsável por preprocessar um programa de entrada, gerando caracteres que marcam 
 * o início e término de blocoscom base na identação. 
 * Facilita o trabalho do parser que pode assumir que existem caracteres especiais 
 * demarcando início e término dos blocos.
 * @author stagni
 *
 */
public class WhitespacePreprocesserReader extends Reader {
	
	
	
	private final int indentChar = '{';
	private final int dedentChar = '}';	
	private static final int NO_CHAR = -2;
	
	private final Reader reader;
	private int currIndentation;
	private Stack<Integer> indentStack;	
	private int leadingChar;
	private int nIndent;
	private int nDedent;
	private boolean waitingNewLine;
	private int currRow;
	private int currCol;
	private boolean endReached;
	public WhitespacePreprocesserReader(Reader reader) {
		this.reader = reader;
		this.currIndentation = 0;
		this.indentStack = new Stack<Integer>();
		this.leadingChar = NO_CHAR;
		this.nIndent = this.nDedent = 0;
		this.waitingNewLine = false;	
		this.currRow = 1;
		this.currCol = 1;
		indentStack.push(0);
	}
	
	
	
	
	public int read(char[] cbuf, int off, int len) throws IOException {
		int k = 0;
		int c;
		
		if (endReached) return -1;
		
		ITERATING:	while (k < len) {
			WRITING_CHAR: do {
				if (nIndent > 0) { c = indentChar; nIndent--; }
				else if (nDedent > 0) { c = dedentChar; nDedent--; }
				else if (waitingNewLine) { c = '\n'; waitingNewLine = false; continue WRITING_CHAR;}
				else if (currIndentation > 0) { c = ' '; currIndentation--; }
				else if (leadingChar != NO_CHAR) { c = leadingChar; leadingChar = NO_CHAR; }
				else { c = readChar(); /*HACK */	if (c == -1) c = '\n'; }
				/* Temos o caracter atual em c */
				if (c == '\n') {
					waitingNewLine = true;
					while ((leadingChar = readChar()) == ' ') currIndentation++;
					if (leadingChar == '\n') continue ITERATING; // Linha em branco: Nao considerar para indentação
					if (leadingChar == -1) currIndentation = 0; // Linha em branco terminada por EOF. Considerar totalmente nao indentada.
					computeNumberOfIndentOrDedent(currIndentation);
					continue ITERATING;
				}
			} while(false);
			// Escreve caracter
			if (c == -1) { this.endReached = true; break ITERATING; }
			cbuf[off + k] = (char)c;
			k++;
		}		
		if (k > 0) return k;
		return -1;
	}
	
	int readChar() throws IOException {
		int c = reader.read();
		if (c == '\t') throw new InvalidCharacterException('\t', currRow, currCol);
		if (c == '\n') {
			currRow++;
			currCol = 0;
		} else
			currRow++;
		return c;
		
	}
	
	void computeNumberOfIndentOrDedent(int indentation) {
		nIndent = nDedent = 0;
		if (indentation > indentStack.peek()) { 
			indentStack.push(indentation);
			nIndent = 1;
		}
		else if (indentation < indentStack.peek()) {
			do {
				indentStack.pop();
				nDedent++;				
			} while(indentation < indentStack.peek());
			if (indentation != indentStack.peek())
				throw new BadDedentException(currRow);
		}
	}

	@Override
	public void close() throws IOException {
		reader.close();
	}
	
	public static class BadDedenting extends RuntimeException {
		private static final long serialVersionUID = 6427493877053188149L;
		public BadDedenting() {
			
		}
		
	}
	
	
	
	public static class InvalidCharacterException extends RuntimeException {
		private static final long serialVersionUID = -3717268111492243815L;
		public final int row;
		public final int col;
		public InvalidCharacterException(char c, int row, int col) {			
			super("Invalid Character [" + c + "] at line " + row + " column " + col);
			this.row = row;
			this.col = col;
		}
	}
	
	public static class BadDedentException extends RuntimeException {
		private static final long serialVersionUID = -3717268111492243815L;
		public final int row;
		public BadDedentException(int row) {			
			super("Bad dedent near line " + row);
			this.row = row;
		}
	}
	
	public char indentChar() {
		return (char)indentChar;
	}
	
	public char dedentChar() {
		return (char)dedentChar;
	}

}
