package br.usp.ime.mac499.xba.transformer;

import static org.objectweb.asm.Opcodes.ACC_STATIC;
import static org.objectweb.asm.Opcodes.ASM4;

import java.util.ArrayList;
import java.util.List;

import org.objectweb.asm.AnnotationVisitor;
import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Type;

/**
 * Le uma classe Java e extrai informações das anotações para criação de uma classe XBA.
 * @author stagni
 *
 */
public class XBAInfoReader extends ClassVisitor {
	private List<XBAMethodInfo> methodsInfo;
	private String className;
	public XBAInfoReader(ClassVisitor cv) {
		super(ASM4, cv);
		methodsInfo = new ArrayList<XBAMethodInfo>();
	}
	
	public void addNewXBAMethodInfo(XBAMethodInfo methodInfo) {
		methodsInfo.add(methodInfo);
	}
	
	@Override
	public void visit(int version, int access, String name, String signature, String superName, String[] interfaces) {
		this.className = name.substring(0, name.lastIndexOf('/') + 1);
		cv.visit(version, access, name, signature, superName, interfaces);
	}
	
	
	@Override 
	public MethodVisitor visitMethod(int access, String name, String desc, String signature, String[] exceptions) { 
		final XBAMethodInfo methodInfo = new XBAMethodInfo(name, desc, (access & ACC_STATIC) != 0);
		MethodVisitor mv = cv.visitMethod(access, name, desc, signature, exceptions);
		return new MethodVisitor(ASM4, mv) {
			private boolean isXBA = false;
			@Override
			public AnnotationVisitor visitAnnotation(String desc, boolean visible) {
				AnnotationVisitor av = mv.visitAnnotation(desc, visible);
				if (desc.equals(Type.getDescriptor(XBAMethod.class))) { 
					av = new AnnotationVisitor(ASM4, av) {
						@Override 
						public AnnotationVisitor visitArray(String name) {
							if (name.equals("selectors")) {
								return new AnnotationVisitor(ASM4, av.visitArray(name)) {
									@Override
									public void visit(String name, Object value) {
										methodInfo.addSelector((String)value);
										av.visit(name,  value);
									}
								};
							}
							return null; // TODO: soltar excessao
						}
					};
					isXBA = true;
				}
				return av;
			}
			
			@Override
			public void visitEnd() {
				if (isXBA) {
					if (Type.getType(methodInfo.descriptor).getArgumentTypes().length != methodInfo.nSelectors())
						throw new IllegalXBAClass("Method " + methodInfo.name + " must have number of selectors equals to number of arguments");
					addNewXBAMethodInfo(methodInfo);
				}
				mv.visitEnd();
			}
		};
	}
	
	@Override 
	public AnnotationVisitor visitAnnotation(String desc, boolean visible) {
		AnnotationVisitor av = cv.visitAnnotation(desc, visible);
		if (desc.equals(Type.getDescriptor(XBAClass.class))) {
			return new AnnotationVisitor(ASM4, av) {
				@Override
				public void visit(String name, Object value) {
					if (name.equals("name") && value instanceof String) 
						className += (String)value;
					av.visit(name, value);
				}
			};
		}
		return av;
	}
	
	@Override 
	public void visitEnd() {
		if (className.endsWith("/"))
			throw new IllegalXBAClass("No XBAClass annotation was found.");
		cv.visitEnd();
	}
	
	public List<XBAMethodInfo> methodsInfo() {
		int n = 0;
		for (XBAMethodInfo methodInfo : methodsInfo) {
			if (methodInfo.isStatic) n++;
		}
		
		if (n != 1) throw new IllegalXBAClass("There must be exactly one static method XBAMethod-annotated");
		
		return this.methodsInfo;
	}

	
	
	public String className() {
		return className;
	}
	
	protected static class XBAMethodInfo {
		public final String name;
		public final String descriptor;
		public final boolean isStatic;
		private String selectorDescriptor;
		private String firstSelector = null;
		private int nSelectors;
		
		public XBAMethodInfo(String name, String descriptor, boolean isStatic) {
			this.name = name;
			this.descriptor = descriptor;
			this.isStatic = isStatic;
			this.selectorDescriptor = "";
		}
		
		public void addSelector(String selector) {
			if (firstSelector == null)
				firstSelector = selector;
			selectorDescriptor += selector + ":";
			nSelectors++;
		}
		
		public int nSelectors() {
			return nSelectors;
		}
		
		public String firstSelector() {
			return firstSelector;
		}
		public String selectorDescriptor() {
			return selectorDescriptor;
		}
	}

	
	
}

class IllegalXBAClass extends RuntimeException {
	public IllegalXBAClass(String message) {
		super(message);
	}
	private static final long serialVersionUID = -8556320242840913390L;
}


