package br.usp.ime.mac499.xba.parser;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.io.FileNotFoundException;
import java.io.UnsupportedEncodingException;

import org.junit.Test;

import br.usp.ime.mac499.xba.ast.ASTBinaryMessage;
import br.usp.ime.mac499.xba.ast.ASTBlock;
import br.usp.ime.mac499.xba.ast.ASTFunctionApplication;
import br.usp.ime.mac499.xba.ast.ASTIdentifier;
import br.usp.ime.mac499.xba.ast.ASTIf;
import br.usp.ime.mac499.xba.ast.ASTInteger;
import br.usp.ime.mac499.xba.ast.ASTLet;
import br.usp.ime.mac499.xba.ast.ASTModule;
import br.usp.ime.mac499.xba.ast.ASTUnaryMessage;
import br.usp.ime.mac499.xba.ast.SimpleNode;

/**
 * Diversos testes verificando que o parser constroi árvores sintáticas corretas 
 * para statements
 * @author stagni
 *
 */
public class ExpressionParserTest {
	XBAFullParser xbaParser;
	
	
	@Test
	public void canParseSimpleExpression() throws UnsupportedEncodingException, FileNotFoundException, ParseException  {
		xbaParser = new XBAFullParser("progs/incomplete/statement/simple-expression");
		SimpleNode node = xbaParser.XBAExpression();
		assertTrue(node instanceof ASTModule);
		assertTrue(node.jjtGetChild(0) instanceof ASTInteger);
		assertEquals(26, ((ASTInteger)node.jjtGetChild(0)).value());
	}
	
	@Test
	public void canParseSimpleMessage()  throws UnsupportedEncodingException, FileNotFoundException, ParseException  {
		xbaParser = new XBAFullParser("progs/incomplete/statement/simple-message");
		SimpleNode node = xbaParser.XBAExpression();
		assertTrue(node instanceof ASTModule);
		assertTrue(node.jjtGetChild(0) instanceof ASTBinaryMessage);
	}
	
	@Test
	public void canParseSimpleMessageWithBlockOperand() throws UnsupportedEncodingException, FileNotFoundException, ParseException  {
		xbaParser = new XBAFullParser("progs/incomplete/statement/simple-message-with-block");
		SimpleNode node = xbaParser.XBAExpression();
		assertTrue(node instanceof ASTModule);
		assertTrue(node.jjtGetChild(0) instanceof ASTBinaryMessage);
		ASTBinaryMessage binMsg = (ASTBinaryMessage)node.jjtGetChild(0);
		assertEquals("then", binMsg.selector());
		assertTrue(binMsg.receiver() instanceof ASTIdentifier);
		assertTrue(binMsg.argument() instanceof ASTBlock);
		
	}
	
	@Test
	public void canParseSimpleMessageWithManyBlocks() throws UnsupportedEncodingException, FileNotFoundException, ParseException  {
		xbaParser = new XBAFullParser("progs/incomplete/statement/simple-message-with-many-blocks");
		SimpleNode node = xbaParser.XBAExpression();
		assertTrue(node instanceof ASTModule);
		assertTrue(node.jjtGetChild(0) instanceof ASTBinaryMessage);
		ASTBinaryMessage elseMsg = (ASTBinaryMessage) node.jjtGetChild(0);
		assertEquals("else", elseMsg.selector());
		assertTrue(elseMsg.argument() instanceof ASTBlock);
		assertTrue(elseMsg.receiver() instanceof ASTBinaryMessage);
		ASTBinaryMessage thenMsg = (ASTBinaryMessage) elseMsg.receiver();
		assertEquals("then", thenMsg.selector());
		assertTrue(thenMsg.argument() instanceof ASTBlock);
	} 
	
	@Test
	public void canParseSimpleAssignment() throws UnsupportedEncodingException, FileNotFoundException, ParseException  {
		xbaParser = new XBAFullParser("progs/incomplete/statement/simple-assignment");
		SimpleNode node = xbaParser.XBAExpression();
		assertTrue(node instanceof ASTModule);
		assertTrue(node.jjtGetChild(0) instanceof ASTBinaryMessage);
		ASTBinaryMessage assignment = (ASTBinaryMessage)node.jjtGetChild(0);
		assertTrue(assignment.receiver() instanceof ASTIdentifier);
		assertTrue(assignment.argument() instanceof ASTInteger);
		ASTInteger number = (ASTInteger)assignment.argument();
		assertEquals(5, number.value());
	}
	
	@Test
	public void canParseAssignmentWithExpression() throws UnsupportedEncodingException, FileNotFoundException, ParseException  {
		xbaParser = new XBAFullParser("progs/incomplete/statement/assignment-with-expression");
		SimpleNode node = xbaParser.XBAExpression();
		assertTrue(node instanceof ASTModule);
		assertTrue(node.jjtGetChild(0) instanceof ASTBinaryMessage);
	}
	
	@Test
	public void canParseIfSimpleExpression() throws UnsupportedEncodingException, FileNotFoundException, ParseException  {
		xbaParser = new XBAFullParser("progs/incomplete/statement/if-expression-simple");
		SimpleNode node = xbaParser.XBAExpression();
		assertTrue(node instanceof ASTModule);
		assertTrue(node.jjtGetChild(0) instanceof ASTIf);
	}
	
	@Test
	public void canParseIfBlockExpression() throws UnsupportedEncodingException, FileNotFoundException, ParseException  {
		xbaParser = new XBAFullParser("progs/incomplete/statement/if-expression-with-block");
		SimpleNode node = xbaParser.XBAExpression();
		assertTrue(node instanceof ASTModule);
		assertTrue(node.jjtGetChild(0) instanceof ASTIf);
	}
	
	@Test
	public void canParseUnaryMessage() throws UnsupportedEncodingException, FileNotFoundException, ParseException  {
		xbaParser = new XBAFullParser("progs/incomplete/statement/unary-message");
		SimpleNode node = xbaParser.XBAExpression();
		assertTrue(node instanceof ASTModule);
		assertTrue(node.jjtGetChild(0) instanceof ASTUnaryMessage);
	}
	
	@Test
	public void canParseSimpleLetExpression() throws UnsupportedEncodingException, FileNotFoundException, ParseException  {
		xbaParser = new XBAFullParser("progs/incomplete/statement/let-expression-simple");
		SimpleNode node = xbaParser.XBAExpression();
		assertTrue(node instanceof ASTModule);
		assertTrue(node.jjtGetChild(0) instanceof ASTLet);
	}
	
	@Test
	public void canParseLetExpressionWithBlock()throws UnsupportedEncodingException, FileNotFoundException, ParseException  {
		xbaParser = new XBAFullParser("progs/incomplete/statement/let-expression-with-block");
		SimpleNode node = xbaParser.XBAExpression();
		assertTrue(node instanceof ASTModule);
		assertTrue(node.jjtGetChild(0) instanceof ASTLet);
	}
	
	@Test
	public void canParseFunctionApplication()throws UnsupportedEncodingException, FileNotFoundException, ParseException  {
		xbaParser = new XBAFullParser("progs/incomplete/statement/function-application");
		SimpleNode node = xbaParser.XBAExpression();
		assertTrue(node instanceof ASTModule);
		assertTrue(node.jjtGetChild(0) instanceof ASTFunctionApplication);
	}
	
	
}
